classdef basic < hgsetget
% Abstract class for creating FileBrowser plugins
%
% fb_abstract.basic Properties:
%   Name - first property
%
% fb_abstract.basic Methods:
%   register_plot - 
%   can_tag - 
%
% only public methods will show a help text
% see "doc fb_abstract.basic"

    %% default properties that can be uninitialized
    properties (SetAccess = protected, GetAccess = public)
        % Name of the plugin
        Name = 'some plugin';
        % Short description of what the plugin is doing
        Description = 'doing something';
        % Who wrote the plugin?
        Author = 'Andreas Kotowicz';
        % Default color for the axis
        DefaultAxisColor = [0 0 0];
        % is this an overlay plot?
        Overlay = 0;
        % in case we want to sort the plugins, we will do so by giving them
        % Order Numbers.
        Order = 0;
        % standard region of interest to start with
        ROI = [1 10000];
        % ROI vector holding the sample indices
        roi = 1:10000;
        % data array holding the processed data from transform_function()
        % this data array will always have the size of the ROI
        transformed_data = [];
        % ignore loading and displaying of new data.
        KEEP_PLOT = 0;
    end
    
    %% default properties that should NOT be saved
    properties (SetAccess = protected, GetAccess = public, Transient = true)
        % FileBrowser window handle
        FBMainWindow = [];
        % in case the plotting functionality of the plugin is used, then we
        % store the main plot handle inside of Handle
        PluginHandle
        % handle which groups all possible plotting handles of the plugin
        PluginHandleGroup
        % what is the handle id of the parent axis that we plot to?
        ParentAxis
        % did the plugin test ok?
        % -1: no test did run yet;  'PLUGIN_TEST_NEVER'
        %  0: test failed;          'PLUGIN_TEST_FAILED'
        %  1: test was ok.          'PLUGIN_TEST_OK'
        testok = -1;
        % Plugin Setup in progress? see 'setup_plugin'
        PLUGIN_SETUP_STATUS = 0;
        % see PLOT_AUTOSCALE_INITIAL below. indicate whether
        % PLOT_AUTOSCALE_INITIAL already ran
        PLOT_AUTOSCALE_INITIAL_EXECUTED = 0;
    end
    

    %% abstract properties that need to be explicitly defined
    properties (Abstract = true, Constant = true)
        % we distinguish plugin classes by it's PLOT_TYPE
        PLOT_TYPE
        % setting up the plot, should the y-limits get autoscaled?
        % this is an intended behaviour for line plots, but not for
        % spectrograms.
        PLOT_AUTOSCALE_INITIAL
    end
    
    %% public properties - user can change them on the command line
    properties
        % reasonable default scanrate that will be changed after file
        % loading - needs public GetAccess, because we access this property
        % in the tag plugin.
        % see 'set.Scanrate()' and 'Scanrate_set()' functions.
        Scanrate = 32000;        
        
        % by default we always show the first channel
        Channel = 1;
        
        % all data can be scaled if requested
        % offset data (add offset) by this value
        y_offset = 0;
        % multiply data by this value
        y_multiply = 1;
        
        % scale data before or after the transform function?
        ScaleBefore = 0;
        
        % should all of the data be transformed at once and we only update
        % the ROI? By default the transformation will only be run on the
        % ROI. But if files are not too long, then it makes sense to
        % transform the data once and only show the correct ROI instead of
        % recomputing the transformation over and over again.
        TRANSFORM_ENTIRE_DATA = 0;
    end

    %% public hidden properties
    properties (SetAccess = public, GetAccess = public, Hidden = true)
        % turn debugging statements on / off
        debug = 0;
        % keep copy of currently selected channel locally with the plugin -
        % this gives great performance for quick browsing of huge files.
        KEEP_DATA_PLUGIN = 0;
        % same as above, but if a new file is loaded, the old data still
        % stays with the plugin
        KEEP_DATA_PLUGIN_IGNORE_FILE_LOADING = 0;
    end
    
    %% private properties
    properties (SetAccess = protected, GetAccess = protected)
        % dummy data used to test the plugin and generate the initial
        % graphs. We should only generate the data once and use it until
        % the plugin is initialized.
        TEST_DATA = [];
        
        % does this plugin require / allow one or multiple data channels?
        MULTIDIM_DATA = 0;
        
        % is a plugin ready to handle entire data transformations?
        TRANSFORM_ENTIRE_DATA_CAPABLE = 0;
        TRANSFORM_DATA_AVAILABLE = 0;
        
        % what are the valid channels in the file?
        ValidChannels = 1;
        
        % file information
        BirdName = '';
        Subdirectory = '';
        FileName = '';
        FILESIZE = '';
        
        % which mode is the plugin being used in?
        % this is important for methods like 'data_set(obj, data)' and 
        % 'Scanrate_set(obj, scanrate)'
        %
        % 
        % 0 = command line processing; obj.transform_function(data); PLUGIN_MODE_COMMAND_LINE
        % 1 = Demo Mode; obj.demo(); PLUGIN_MODE_DEMO_MODE
        % 2 = FileBrowser; obj.setup(); PLUGIN_MODE_FILEBROWSER
        PLUGIN_MODE = 0;
        
        % indicate whether 'set_dataset_info' is being called, so we don't
        % run certain methods.
        PLUGIN_DATASET_INFO_STATE = 0;
        
        % indicates whether plot() is called from update_plot() or not.
        UPDATE_PLOT = 0;
        
        % does the plugin require a data channel? some plugins run on
        % external data and therefore don't need a data channel.
        BASIC_NEEDS_CHANNEL = 1;
        BASIC_CREATE_DEFAULT_CONTEXT_MENU = 1;
    end
    
    %% CONSTANT PROPERTIES used to read out config states
    properties (Constant = true)
        PLUGIN_MODE_COMMAND_LINE = 0;
        PLUGIN_MODE_DEMO_MODE = 1;
        PLUGIN_MODE_FILEBROWSER = 2;
        PLUGIN_TEST_NEVER = -1;
        PLUGIN_TEST_FAILED = 0;
        PLUGIN_TEST_OK = 1;
        PLUGIN_SETUP_IN_PROGRESS = 1;
        PLUGIN_SETUP_DONE = 0;
        PLUGIN_DATASET_INFO_RUNNING = 1;
        PLUGIN_DATASET_INFO_NOT_RUNNING = 0;
    end

    %% private properties that should NOT be saved
    properties (SetAccess = protected, GetAccess = protected, Transient = true)
        % instance of FileBrowser client object that we use so our plugin
        % can interact with FB and FL. This object is unique for every
        % plugin instance that is created on the command line.
        % FB_GUI_object is the same if plugins are created through file
        % browser.
        FB_GUI_object
        
        % event listener
        ListenerBirdChange
        ListenerFileChangePrepare
        ListenerFileChange
        ListenerROIChange
        ListenerKeyPress
        ListenerHidePlugin
        
        % context menu for each plot / overlay
        ContextMenu
        
        % context menu for the axis
        ContextMenuAxis
        AddPluginsMenu
        
        % context menu entries
        %
        % Name of plugin shown in context menu
        ContextName
        
        % individual entries below the name
        ContextScaleBeforeAfter
        ContextTransformEntireFile
        ContextChannel
        ContextValidChannelsHeader
        ContextValidChannels
        ContextScanrate
        ContextResolution
        ContextReRunRequired
        ContextPlugins
        ContextRegisteredPluginsDelete
        ContextRegisteredPluginsShowHide
        ContextRegisteredPluginsReRunPluginRequired
        ContextRegisteredPlugins
    end
    
    
    %% properties that need to be calculated / read out each time
    properties (Dependent = true, SetAccess = protected, GetAccess = protected)
        % registry for plugins
        plugins
        % registry of axes objects
        axes_objects
    end    

    properties (Dependent = true, SetAccess = private, GetAccess = public)
        % if you want to overwrite data (only reasonable in scripting / 
        % demo mode), check "data_set()".
        data
    end
    
    
    properties (SetAccess = public, GetAccess = public, Hidden = true, Transient = true)
        internal_data
    end
    
    %% events
    events
        HidePlugin
        ShowPlugin
        % event which notifies subhandles that the plugin is about to be
        % deleted
        DeletePlugin
        % notify plugin subcomponents that a new file has been loaded.
        NewFile
        % notify plugin subcomponents that we use a new ROI
        NewROI
    end    
    
    %% required methods - an instance of a class can not be created if
    % these are not defined!
    methods (Abstract = true, Access = protected, Hidden = true)
        plugin_setup_plot(obj, Tag);
        plugin_custom_plot(obj, data);
        plugin_createContextMenuPlugin(obj);
        class_plot_routine(obj, data);
        class_hide_handles(obj, src, evnt);
        class_show_handles(obj, src, evnt);
        timepoints = class_return_samplenumbers(obj);
        % plugin_threshold_function(obj);
    end
    
    %% default object methods that always exist and that can be redefined
    methods
        
        %% this class has no DELETE method! if you want to delete the
        % object, call ContextFuncDeleteAxis()
        % the reason for this is that a plugin is not one, but actually two
        % objects: 1) axis object, 2) plotting object.
        %
        
        
        %% SET AND GET METHODS - need to be in public scope

        function set.Scanrate(obj, Scanrate)
        % !!!!!!!!!!!!!!!!!!      PLEASE BE CAREFUL     !!!!!!!!!!!!!!!!!!!!!!!!!!!
        %
        %  Setting the 'Scanrate' might have serious consequences for a transform
        %  function, because, e.g., a pre-calculated filter kernel might depend on
        %  the 'Scanrate'. The current version of the plugins tries to
        %  check for these dependencies as good as possible, however, if
        %  you feel like things don't work as expected, please contact me.
        %
        % !!!!!!!!!!!!!!!!!!      PLEASE BE CAREFUL     !!!!!!!!!!!!!!!!!!!!!!!!!!!
        %

            
            
            % Plugin used with FileBrowser, user trying to overwrite
            % Scanrate -> not allowed.
            if chk_plugin_bound_to_filebrowser_scanrate_set_from_commandline(obj)
                disp('This plugin is coupled to a FileBrowser session - you can''t overwrite the ''Scanrate'' manually.');
                return;
            end
            
            obj.chk_pos_scalar(Scanrate, 'Scanrate');
            prev_scanrate = obj.Scanrate;
            
            % if the scanrate is the same as before, don't trigger any
            % updates.
            if prev_scanrate == Scanrate
                return;
            end
            
            % new scanrate
            obj.Scanrate = Scanrate;
            
            % re-run 'load_required' because this plugin might depend on 'Scanrate'.
            load_required(obj);

            % Don't run update_plot() below if 'Scanrate' is set from
            % 'set_dataset_info()', which is only possible in
            % 'PLUGIN_MODE_FILEBROWSER' mode. The plotting will be
            % initiated later (see fb_fl_from_external ~ line 190)
            if chk_plugin_bound_to_filebrowser_scanrate_set_from_dataset_info(obj)
                return;
            end            
            
            % COMMANDLINE MODE -> nothing to plot -> we are done.
            if chk_plugin_mode_command_line(obj)
                return;
            end
            
            % 'PLUGIN_MODE_DEMO_MODE' - write Scanrate back to fake
            % 'FB_GUI_object' object & call 'update_plot()'
            chk_plugin_mode_demo_scanrate_set_from_commandline(obj)

        end
        
        function set.Channel(obj, Channel)
            % ContextFuncChannel (which calls this function) will call
            % update_plot once this method is done.
            
            if obj.MULTIDIM_DATA == 0
                message = [' > Plugin "' obj.Name '" expects a single channel.'];
                obj.chk_pos_scalar(Channel, 'Channel', message);
            else
                message = [' > Plugin "' obj.Name '" expects multiple channels.'];
                obj.chk_pos_array(Channel, 'Channel', message);
            end
                           
            obj.Channel = Channel;
            ContextUpdateChannel(obj);
            
            % make sure to update the internal data in case the channel has
            % changed
            if obj.KEEP_DATA_PLUGIN == 1 || obj.KEEP_DATA_PLUGIN_IGNORE_FILE_LOADING == 1
                update_internal_data(obj, 1);
            end
            
            if obj.TRANSFORM_ENTIRE_DATA == 1
                obj.TRANSFORM_DATA_AVAILABLE = 0;
                update_plot(obj);
            end
            
        end
        
        function set.y_offset(obj, y_offset)
            obj.y_offset = y_offset;
            obj.TRANSFORM_DATA_AVAILABLE = 0;
            overwrite_data_handle_obj(obj);
        end
        
        function set.y_multiply(obj, y_multiply)
            obj.y_multiply = y_multiply;
            obj.TRANSFORM_DATA_AVAILABLE = 0;
            overwrite_data_handle_obj(obj);
        end
        
        function set.plugins(obj, plugins)
            setappdata(obj.FBMainWindow, 'plugins', plugins);
        end

        function set.axes_objects(obj, axes_objects)
            setappdata(obj.FBMainWindow, 'axes_objects', axes_objects);
        end
        
        function set.TRANSFORM_ENTIRE_DATA(obj, new_value)
            
            % what's the new value
            if new_value == 1
                new_checked = 'on';
            else
                new_checked = 'off';
                obj.TRANSFORM_DATA_AVAILABLE = 0;
            end
            
            set(obj.ContextTransformEntireFile, 'Checked', new_checked);
            obj.TRANSFORM_ENTIRE_DATA = new_value;
            
        end
        
        function set.ROI(obj, new_ROI)
            roi_vector = new_ROI(1) : new_ROI(2);
            obj.roi = roi_vector;
            obj.ROI = new_ROI;            
        end
        
        function plugins = get.plugins(obj)
            if ishandle(obj.FBMainWindow)
                plugins = getappdata(obj.FBMainWindow, 'plugins');
            else
                plugins = [];
            end
        end
        
        function axes_objects = get.axes_objects(obj)
            if ishandle(obj.FBMainWindow)
                axes_objects = getappdata(obj.FBMainWindow, 'axes_objects');
            else
                axes_objects = [];
            end
        end        
        
        function returned_data = get.data(obj)

            if obj.KEEP_DATA_PLUGIN == 1 || obj.KEEP_DATA_PLUGIN_IGNORE_FILE_LOADING == 1
                returned_data = obj.internal_data;
                return;
            end
            
            returned_data = return_data(obj);
            if isempty(returned_data)
                keyboard;
            end
        end
        
        function set.KEEP_DATA_PLUGIN(obj, keep_me)
            % only 0 and 1 and valid modes
            if keep_me == 1 || keep_me == 0
                update_internal_data(obj, keep_me);
                obj.KEEP_DATA_PLUGIN = keep_me;
            end
        end
        
        function set.KEEP_DATA_PLUGIN_IGNORE_FILE_LOADING(obj, keep_me)
            % only 0 and 1 and valid modes
            if keep_me == 1 || keep_me == 0
                update_internal_data(obj, keep_me);
                obj.KEEP_DATA_PLUGIN_IGNORE_FILE_LOADING = keep_me;
            end
        end        
        
        %% allow read only access to data 
        % since data's 'SetAccess' is set to private, this error message
        % will never be shown! There this is more for documentation
        % purposes.
        function set.data(obj, data) %#ok<MANU,INUSD>
            disp('"data" is read-only -- your input is ignored.');
            disp('if you really want to overwrite data (only reasonable in scripting / demo mode), check "data_set"');
            return;
        end
        
        %% disp function for matlab 2008a (and possibly higher versions)
        % seems like some matlab implementations crash if we don't provide
        % our own disp() method.
        function disp(obj)
            a = class(obj);
            fprintf('  FileBrowser plugin: <a href = "matlab:help(''%s'')">%s</a>\n', a, a);
            fprintf('   <a href = "matlab:properties(''%s'')">properties</a>;', a);
            fprintf(' <a href = "matlab:methods(''%s'')">methods</a>;', a);
            fprintf(' <a href = "matlab:events(''%s'')">events</a>;', a);
            fprintf(' <a href = "matlab:superclasses(''%s'')">superclasses</a>;', a);
            fprintf(' <a href = "matlab:fb_inspect_plugin(''%s'')">inspect class</a>;\n', a);
        end
        
    end

    
    %% you'll find only methods' signatures below this point.
    
    %% HIDDEN methods
    methods (Access = protected, Hidden = true)
        
        %% registers new object & finalizes plugin
        register_plot(obj)
        register_overlay(obj)
        
        %% THESE FUNCTIONS are place holder and are redefined by each
        % plugin as needed

        plugin_transform_function(obj, Data) % bla bla
        plugin_finalize_GUI(obj)
        plugin_keypressfunction(obj, evnt)
        data = plugin_threshold_function(obj)
        status = plugin_load_required(obj)
        plugin_file_changed(obj)
        data = plugin_scale_function(obj, data)
        status = plugin_test(obj)
        buffer = plugin_which_buffer(obj, sample)
        
        %%
        setup_plot_pre(obj)
        createListener(obj)        
        finish_plugin(obj)
        registerPluginAxisContextMenu(obj)
        createContextMenuMinimal(obj, handle_to_use)
        
        %% extended data handling
        update_internal_data(obj, mode)
        returned_data = return_data(obj)
        
        %% context setup functions

        createContextMenuPluginMain(obj, handle_to_use)
        ContextSetupName(obj)
        ContextSetupChannel(obj)
        ContextSetupScaleData(obj)        
        ContextSetupScaleBeforeAfter(obj)
        ContextSetupTransformEntireFile(obj)
        ContextSetupTagData(obj)
        ContextSetupScanrate(obj)
        ContextSetupResolution(obj)     
        ContextSetupReRunRequired(obj)
        ContextSetupKeepPlot(obj)
        ContextSetupKeepDataPlugin(obj)
        ContextSetupKeepDataPluginIgnoreFileLoading(obj)
        ContextSetupHide(obj)
        
        %% context update functions
        ContextUpdateChannel(obj)
        ContextUpdateScanrate(obj)
        ContextUpdateResolution(obj)
        ContextUpdateValidChannels(obj)

        %% other context
        ContextGetValidChannels(obj)
        
        %% other stuff
        ShowHide(obj, src, evnt, status)
        status = check_Visiblity(obj)        
        status = invert_Visiblity(obj)
        SetupDeletePluginListener(obj, Handle)
        
        register_data_handle(obj)
        register_axes_object(obj, axes_object)
        unregister_axes_object(obj, axes_object)
        overwrite_data_handle_obj(obj)
        update_ROI(obj)
        update_y_ticks(obj)
        set_dataset_info(obj)
        demo_mode_update_info_new_data(obj, nbr_samples_per_channel)
        
    end

    %% CALLBACKS, LISTENERS and CONTEXTMENUs are hidden
    methods (Hidden = true)
        
        % callbacks - they need to be public!
        listenChangedBird(obj, src, evnt)
        listenChangedFile(obj, src, evnt)
        listenChangedFilePrepare(obj, src, evnt)
        listenChangedROI(obj, src, evnt)
        listenKeyPress(obj, src, evnt)
        ListenHidePlugin(obj, src, evnt)
        ListenDeleteObject(obj, src, evnt, ObjectToDelete)
        
        ContextFuncChannel(obj, src, evnt, OldChannel, NewChannel)
        ContextFuncScaleBeforeAfter(obj, src, evnt)
        ContextFuncTransformEntireFile(obj, src, evnt)
        ContextFuncResolution(obj, src, evnt)
        ContextFuncReRunRequired(obj, src, evnt)
        ContextFuncChangeColor(obj, src, evnt, Handle)
        ContextFuncChangeLineWidth(obj, src, evnt, Handle)
        ContextFuncShowHorizontalLine(obj, src, evnt, initial_position, srcTag)
        ContextFuncHideHorizontalLine(obj, src, evnt, srcTag)
        ContextFuncAddPlugin(obj, src, evnt, plugin, nbr_copies, extra_argument)
        ContextFuncReloadPluginList(obj, src, evnt)
        ContextFuncShowHidePlugin(obj, src, evnt)
        ContextFuncDeletePlugin(obj, src, evnt)
        ContextFuncDeleteAxis(obj, src, evnt)
        ContextFuncAxisSwitchScaleMode(obj, src, evnt)
        ContextFuncAxisChangeLimitsY(obj, src, evnt)
        ContextFuncAxisChangeLimitsX(obj, src, evnt)
        ContextFuncAxisShowYTicks(obj, src, evnt)
        ContextFuncScaleData(obj, src, evnt)
        ContextFuncInspectPlugin(obj, src, evnt)
        context_menu_line = createContextMenuLine(obj, windowHandle, lineHandle, context_menu_line)
        
    end
    
    %% PLUGIN INTERFACE METHODS
    methods (Sealed = true, Access = protected, Hidden = true)
        
        % automatic setup method for OVERLAYS - an axis already exists
        setup_overlay(obj, FBMainWindow, FB_GUI_object, ParentAxis, channel_to_use, PLUGIN_MODE)
        setup_plugin(obj, Tag, channel_to_use)
        setup_axis(obj, varargin)
        createContextMenuAxis(obj)
        createContextMenuAxisEntries(obj)
        create_plugin_menu(obj)
        create_plugin_list(obj)
        setup_plot(obj, Tag)
        custom_plot(obj, data)
        createContextMenuPlugin(obj)
        plot_routine(obj, data)

    end    
    
    %% PUBLIC INTERFACE METHODS - CAN NOT BE REDEFINED
    methods (Sealed = true)

        %% automatic setup method for PLOTS
        setup(obj, FBMainWindow, FB_GUI_object, DataPanel, bottom, height, NewNbrAxis, Tag)

        %% demo method which shows what the plugin is doing.
        demo(obj, FB_GUI_object)

        %% tests the plugin
        test(obj)

        %% default plot function
        plot(obj, data)

        %% the default transform_function returns the data as is
        Data = transform_function(obj, Data)

        %% threshold function
        data = threshold_function(obj)

        %% scaling function
        data = scale_transform_data(obj, data)

        %% allow user to define a function that can load any required data
        status = load_required(obj)

        %% interface to re-run plotting
        update_plot(obj, update_y_scaling)
        
        %% return the timepoints (in scanrate) for each y value
        timepoints = return_samplenumbers(obj);
        
        %% return the timepoints (in scanrate) for each y value
        buffer = which_buffer(obj, sample);
        
        %% interface for doing additional GUI stuff once plugins are fully
        % initialized
        finalize_plugin_GUI(obj)
        
        %% Inidicate that a new file has been loaded
        file_changed(obj)

        %% hide the plugin handle(s)
        hide_plugin(obj, src, evnt)

        %% show the plugin handle(s)
        show_plugin(obj, src, evnt)

        %% tell whether this plugin is tagging capable
        CanTag = can_tag(obj)

        %% determine whether plugin is currently used for plotting
        isPlot = is_plot(obj);
        
        %% create some stats
        gen_stats_usage(obj);
        
        %% set the x - limits of the parent axis object
        set_xlims(obj, data_length)

        %% set the x limits using the current ROI.
        set_xlims_roi(obj)
        
        %% set the y limits using the data range
        set_ylims_data(obj)
        
        %% special function for PLUGIN_MODE_DEMO_MODE only!!
        data_set(obj, data)
        
        %% deprecated - you can set the 'Scanrate' now directly!
        Scanrate_set(obj, scanrate)
        
        %%
        [rerunpluginrequired, deleteplugin, showhideplugin] = return_axis_menu_handles(obj)
        
        %% check whether plugin has been tested.
        status = chk_plugin_not_tested(obj)
        
        %% check whether plugin is in PLUGIN_MODE_FILEBROWSER mode and user is trying to set 'Scanrate' manually
        status = chk_plugin_bound_to_filebrowser_scanrate_set_from_commandline(obj)
        status = chk_plugin_bound_to_filebrowser_scanrate_set_from_dataset_info(obj)
        status = chk_plugin_mode_demo(obj)
        status = chk_plugin_mode_command_line(obj)
        chk_plugin_mode_demo_scanrate_set_from_commandline(obj)
        
    end

    %% Helperfunctions (independent) of object
    methods (Static = true)
        chk_pos_scalar(value, varname, message)
        chk_pos_array(value, varname, message)
        show_error(message, filename, entire_stack, object_name, function_name, identifier)
        data = gen_demo_data(use_real_data)
        tf = isaPropertyOf(anyobj,PropName)
        [data, nbr_channels, nbr_samples_per_channel] = chk_data_for_valid_dimensions(data)
    end
    
    
    %% NOT USED loadobj and saveobj methods - we use transient properties
    % for listeners and handles - that takes care of all potential problems
    methods (Static = true, Access = private)
        obj = loadobj_notused(S)
    end
    
    methods (Access = private)
        S = saveobj_notused(obj)
    end
    
end
%% EOF
