classdef plot_buffered < fb_abstract.plot_buffered_common
% this is the super class for buffered plots

    % "PLOT_TYPE" can not be changed under any circumstances!
    properties (Constant = true)
        PLOT_TYPE = 3;
    end
    
    properties
        buffersize = 256;
        nonoverlap = 128;
        samplestart = 128;
    end
    
    properties (Dependent = true)
        % some plugins need the overlap variable.
        overlap
    end
    
    properties (SetAccess = protected, GetAccess = public)
        % number of buffers used for this plot
        nbr_buffers
    end
    
    
    %% public methods - none. We do not provide a constructor, because this
    % is an abstract class.
    methods
        
        % calculate the overlap if needed
        function overlap = get.overlap(obj)
            overlap = obj.buffersize - obj.nonoverlap;
            
            overlap_tmp = get_overlap(obj);
            if isempty(overlap_tmp)
                return;
            end
            
            overlap = overlap_tmp;
        end
        
        % calculate the buffersize if needed
        function buffersize = get.buffersize(obj)
            buffersize = obj.buffersize;
            
            buffersize_tmp = get_buffersize(obj);
            if isempty(buffersize_tmp)
                return;
            end
            
            buffersize = buffersize_tmp;
        end
        
        % calculate the nonoverlap if needed
        function nonoverlap = get.nonoverlap(obj)
            nonoverlap = obj.nonoverlap;
            
            nonoverlap_tmp = get_nonoverlap(obj);
            if isempty(nonoverlap_tmp)
                return;
            end
            
            nonoverlap = nonoverlap_tmp;
        end
        
        % custom get methods - can be overwritten
        function overlap = get_overlap(obj) %#ok<MANU>
            overlap = [];
        end
        
        function nonoverlap = get_nonoverlap(obj) %#ok<MANU>
            nonoverlap = [];
        end
        
        function buffersize = get_buffersize(obj) %#ok<MANU>
            buffersize = [];
        end
        
    end
       
    %% hidden methods
    methods (Access = protected, Hidden = true)
        
        %% plotting function assumes that data is in buffers
        function plugin_custom_plot(obj, data)
            obj.create_buffered_plot(data, obj.Scanrate, obj.samplestart, obj.nonoverlap, obj.PluginHandle, obj.shift_buffers_by);
        end
        
        %%
        function buffer = plugin_which_buffer(obj, sample)
            
            if sample <= obj.buffersize
                buffer = 1;
            else
                buffer = ceil((sample - obj.buffersize) / obj.nonoverlap) + 1;
            end
            % above formula is based on the following observations:
            %
            % (1)
            % buffersize = 512
            % nonoverlap = 32
            % 1    -  512: 1
            % 513  -  544: 2
            % 545  -  576: 3
            % 577  -  608: 4
            % 609  -  640: 5
            %
            % (2)
            % buffersize = 100
            % nonoverlap = 100
            % 1    -  100: 1
            % 101  -  200: 2
            % 201  -  300: 3
            % 301  -  400: 4
            % 401  -  500: 5
            
        end
        
    end
    
    %% STATIC METHODS THAT CAN BE REUSED BY OTHER CLASSES
    methods (Static = true)
        
        function create_buffered_plot(data, scanrate, samplestart, nonoverlap, handle, shift_buffers_by)
            
           %% transfer each buffer-onset-x-location into time.
            nbr_buffers = size(data, 2); %#ok<PROP>
            
            % this is equal to calling obj.transfer_buffers_to_time(nbr_buffers)
            % but static methods can't call object methods
            x_buffered = (samplestart + ((((1:nbr_buffers) + shift_buffers_by) - 1) * nonoverlap)) / scanrate; %#ok<PROP>

            %% check if dimensions are ok
            if size(x_buffered, 2) ~= size(data, 2)
                disp('Can not plot data: X and Y vectors do not have the same dimensions.');
                return;
            end
            
            %% plot the data in units of seconds
            set(handle, 'XData', x_buffered, 'YData', data);
        end
       
        function nbr_buffers = calc_num_buffs(nbr_samples, samplestart, nonoverlap)
            nbr_buffers = 1 + floor((nbr_samples - samplestart) / nonoverlap);
        end
        
    end
    
    
    %% API
    methods (Access = public, Hidden = true)
        
        function calc_number_buffers(obj, data)
            % let's ignore the entire ROI thing here. The only class that
            % is / was using calc_number_buffers (with the ROI) is
            % "plot_buffered_with_time_points" - which is not used anymore.
            %             size_ROI = size(obj.roi, 2);
            %             size_data = 
            %             % if user does:
            %             % $ a = fb_plugin_RMS()
            %             % $ a.transform_function(rand(1,512))
            %             % then ROI is not 512, since it's ignored.
            %             % therefore I need this manual check here.
            %             if size_ROI > size_data
            %                 nbr_samples = size_data;
            %             else
            %                 nbr_samples = size(data(obj.roi), 2);
            %             end
            nbr_samples = size(data, 2);
            obj.nbr_buffers = obj.calc_num_buffs(nbr_samples, obj.samplestart, obj.nonoverlap);
        end
        
        function indices = get_roi_indices_for_buffers(obj, nbr_buffers)
            roi = obj.ROI;
            indices = roi(1) : roi(1) + (nbr_buffers * obj.nonoverlap) - 1 + obj.samplestart;
        end
        
        %% DIFFERENCE BETWEEN buffered_new and buffered
        function time_in_samplingrate = transfer_single_buffer_to_sample(obj, buffer)
            
            % buffer 0 does not exist - only transfer non-zero elements
            non_zero_entries = find(buffer > 0);
            buffer(non_zero_entries) = buffer(non_zero_entries) + obj.shift_buffers_by;
            time_in_samplingrate_non_empty = (obj.samplestart + (buffer(non_zero_entries)-1) * obj.nonoverlap);
            buffer(non_zero_entries) = time_in_samplingrate_non_empty;
            time_in_samplingrate = buffer;
            
        end
        
    end
    
    
end
%% EOF
